import matplotlib.pyplot as plt
import numpy as np

from matplotlib.animation import FuncAnimation

from whirly.fourier import FourierField
from whirly.utils import make_grid

def make_animation(outputs, output_tau, m=None):
    """
    Animate the solution to a two-dimensional PDE.

    Parameters
    ----------
    outputs : [whirly.fourier.FourierField]
        The set of snapshots of the solution, most likely the output of a call
        to the solve method of a PseudospectralSolver subclass.
    output_tau : float
        The time step between snapshots in outputs.
    m : int, optional
        If m is not None, snapshots will be resampled to m grid points.

    Returns
    -------
    animation : matplotlib.animation.FuncAnimation
        The animated solution.

    """

    p = outputs[0].p
    if m is None:
        m = outputs[0].m

    fields = [field.resample(m).real for field in outputs]
    vmax = max(abs(field).max() for field in fields)

    fig, ax = plt.subplots()
    fig.set_size_inches(6, 6)

    zero = FourierField.from_real(np.zeros((m, m)), p)
    _, mesh = plot(zero, ax, vmax)

    def update(i):
        mesh.set_array(fields[i].ravel())
        ax.set_title(f't = {(i * output_tau):.3f}')

        return mesh

    return FuncAnimation(fig, update, np.arange(len(fields)), interval=20)

def plot(field, ax=None, vmax=None):
    """
    Plot a FourierField object on its domain.

    Parameters
    ----------
    field : whirly.fourier.FourierField
        The function to be plotted.
    ax : matplotlib.pyplot.axis, optional
        The axis on which to make the plot. If None, an axis will be created.
    vmax : float, optional
        The maximum value to use for the symmetric colormap. If None, cmax will
        be computed as the infinity norm of field.

    Returns
    -------
    ax : matplotlib.pyplot.axis
        The axis containing the plot. If ax is passed as an argument, the
        returned axis is the same object.
    mesh : matplotlib.collections.QuadMesh
        The result of the call to pcolormesh.

    """

    if ax is None:
        fig, ax = plt.subplots()
        fig.set_size_inches(6, 6)

    ax.set_xlim(0, field.p)
    ax.set_ylim(0, field.p)

    ticks = np.linspace(0, field.p, 6)
    ax.set_xticks(ticks)
    ax.set_yticks(ticks)

    data = field.real
    if vmax is None:
        vmax = abs(data).max()

    x, y = make_grid(field.m, field.p)
    mesh = ax.pcolormesh(
        x, y, data,
        vmin=(-vmax),
        vmax=vmax,
        cmap='RdBu_r',
        shading='nearest'
    )

    return ax, mesh



import numpy as np

# Provide short aliases for Fourier operations, taking real parts as necessary.
fft = np.fft.fft2
ifft = lambda a: np.fft.ifft2(a).real

def make_grid(m, p):
    """Makes a grid with m nodes on the square domain [0, p]^2."""
    x, y = np.mgrid[0:p:((m + 1) * 1j), 0:p:((m + 1) * 1j)]

    return x[:-1, :-1], y[:-1, :-1]

def make_wavenumbers(m, p):
    """Makes arrays of Fourier wavenumbers for m nodes on [0, p]^2."""
    freqs = 2 * np.pi * np.fft.fftfreq(m, d=(p / m))
    ell, k = np.meshgrid(freqs, freqs)

    return k, ell



import numpy as np
import pytest

from whirly.fourier import FourierField
from whirly.utils import make_grid

def make_func(k=1, ell=2):
    def func(x, y):
        return np.sin(k * 2 * np.pi * x) + 0.5 * np.cos(ell * 2 * np.pi * y)

    return func

def test_add_fields():
    a = FourierField.from_func(make_func(), 101)
    b = FourierField.from_func(make_func(2, 1), 101)

    assert np.allclose(a.real + b.real, (a + b).real)

def test_real():
    x, y = make_grid(101, 1)
    real = make_func()(x, y)
    field = FourierField.from_real(real)

    assert np.allclose(real, field.real)

def test_m_must_be_odd():
    hats = np.random.rand(100, 100)
    with pytest.raises(ValueError):
        _ = FourierField(hats, 1)

def test_multiply_fields():
    a = FourierField.from_func(make_func(), 101)
    b = FourierField.from_func(make_func(), 101)

    assert np.allclose(a.real * b.real, (a * b).real)

def test_resample_coarser():
    a = FourierField.from_func(make_func(), 101).resample(75)
    b = FourierField.from_func(make_func(), 75)

    assert np.allclose(a.hats, b.hats)

def test_resample_finer():
    a = FourierField.from_func(make_func(), 101).resample(151)
    b = FourierField.from_func(make_func(), 151)

    assert np.allclose(a.hats, b.hats)



import numpy as np

from whirly.integrators import IFRK4Integrator
from whirly.utils import make_wavenumbers

class PseudospectralSolver:
    """
    Abstract class for pseudospectral solutions to doubly-periodic PDEs.

    Subclasses should
        - define an __init__ method that accepts a time step tau, grid number m,
            and domain size p, and pass all to PseudospectralSolver.__init__
        - define an array L with the coefficient for each Fourier mode in the
            linear part of the PDE
        - define a nonlinear method that accepts a FourierField representing
            the vorticity and returns a FourierField representing the nonlinear
            part of the PDE

    Instances of PseudospectralSolver should not be created directly.

    """

    def __init__(self, tau, m, p):
        """
        Initialize a PseudospectralSolver.

        Note that PseudospectralSolver is meant to be subclassed and instances
        should not be created directly.

        Parameters
        ----------
        tau : float
            The time step.
        m : int
            The number of grid nodes.
        p : float
            The length of a side of the domain.

        """

        self.tau = tau
        self.m = m
        self.p = p

    def solve(self, zeta_initial, T, output_tau=None):
        """
        Solve a nonlinear PDE on a doubly-periodic square domain.

        This method can be used by classes that subclass PseudospectralSolver
        and relies on subclasses implementing nonlinear and defining L.

        Parameters
        ----------
        zeta_initial : whirly.fourier.FourierField
            A FourierField representing the initial solution field.
        T : float
            The final time to solve until.
        output_tau : float
            The time frequency which the solution field should be output. Must
            be smaller than the effective time step, which is calculated by
            adjusting the tau attribute of the instance such that it divides
            evenly into T. If None, the solution will be output every time step.

        Returns
        -------
        solution : [whirly.fourier.FourierField]
            A list of FourierField instances representing the solution field at
            time frequency governed by output_tau

        """

        n_steps = round(T / self.tau)
        tau = T / n_steps

        if output_tau is None:
            output_tau = tau

        skip = round(output_tau / tau)
        zeta = zeta_initial
        outputs = [zeta]

        integrator = IFRK4Integrator(tau, self.L, self.nonlinear)
        for i in range(1, n_steps + 1):
            zeta = integrator.step(zeta)
            if i % skip == 0:
                outputs.append(zeta)

        return outputs

class NavierStokesSolver(PseudospectralSolver):
    """PseudospectralSolver for two-dimensional Navier-Stokes."""

    def __init__(self, tau, m, p, Re):
        """
        Initializes a NavierStokesSolver.

        Parameters
        ----------
        tau : float
            The time step.
        m : int
            The number of grid nodes.
        p : float
            The length of a side of the domain.
        Re : float
            The Reynolds number.

        """

        super().__init__(tau, m, p)

        self.k, self.ell = make_wavenumbers(self.m, self.p)
        K_sq = (self.k ** 2) + (self.ell ** 2)
        self.L = -(1 / Re) * K_sq

        mask = K_sq != 0
        self.inverse_laplacian = np.zeros_like(K_sq)
        self.inverse_laplacian[mask] = -1 / K_sq[mask]

    def nonlinear(self, zeta):
        """
        Calculates the advection of vorticity.

        First, the streamfunction is calculated by solving Laplace's equation,
        and then u and v are found through spectral differentiation and used to
        compute the (negative) advection term.

        Parameters
        ----------
        zeta : whirly.fourier.FourierField
            The current vorticity field.

        Returns
        -------
        advection : whirly.fourier.FourierField
            The nonlinear advection term -u dot grad(zeta).

        """

        psi = self.inverse_laplacian * zeta
        u = -1j * self.ell * psi
        v = 1j * self.k * psi

        zeta_x = 1j * self.k * zeta
        zeta_y = 1j * self.ell * zeta

        return -(u * zeta_x + v * zeta_y)



from functools import partial

import numpy as np

from whirly.fourier import FourierField
from whirly.solvers import NavierStokesSolver

def test_taylor_green():
    m, p = 101, 1
    tau, T = 0.01, 0.5
    nu = 1e-2

    def zeta_func(x, y, t):
        decay = 2 * np.pi * np.exp(-4 * (np.pi ** 2) * nu * t)
        oscillation = np.cos(2 * np.pi * x) + np.cos(2 * np.pi * y)

        return decay * oscillation

    zeta_i = FourierField.from_func(partial(zeta_func, t=0), m, p)
    zeta_f = FourierField.from_func(partial(zeta_func, t=T), m, p)

    solver = NavierStokesSolver(tau, m, p, Re=(1 / nu))
    solution = solver.solve(zeta_i, T)[-1]

    assert np.allclose(zeta_f.real, solution.real)



import numpy as np

class IFRK4Integrator:
    """
    Implementation of fourth-order Runge-Kutta integrating factor time stepping.

    We follow equations (18) and (19) of Yang et al. (2021), implementing an
    integrating factor scheme based on the classic RK4 scheme.

    """

    def __init__(self, tau, L, nonlinear):
        """
        Initializes an integrator given parameters of the ODE system.

        Parameters
        ----------
        tau : float
            The time step. Corresponds to h in Yang et al. (2021).
        L : float or np.ndarray
            The scalar or array of scalars representing the linear part of the
            ODE system.
        nonlinear : callable
            Function returning the nonlinear part of the ODE system. Corresponds
            to N in Yang et al. (2021).

        """

        self.tau, self.nonlinear = tau, nonlinear
        self.E = np.exp(self.tau * L)
        self.H = np.sqrt(self.E)

    def step(self, y):
        """
        Steps an array of function values forward one time step.

        Parameters
        ----------
        y : np.ndarray
            The array of function values at the current time step.

        Returns
        -------
        y_next : np.ndarray
            The array of function values at the next time step.

        """

        N = {1 : self.nonlinear(y)}
        N[2] = self.nonlinear(self.H * (y + 0.5 * self.tau * N[1]))
        N[3] = self.nonlinear(self.H * y + 0.5 * self.tau * N[2])
        N[4] = self.nonlinear(self.E * y + self.tau * self.H * N[3])

        return self.E * y + self.tau * (
            (1 / 6) * self.E * N[1] +
            (1 / 3) * self.H * N[2] +
            (1 / 3) * self.H * N[3] +
            (1 / 6) * N[4]
        )



from itertools import product

import numpy as np

import whirly.utils as utils

class FourierField:
    """
    Array of Fourier coefficients with automatic spectral transforms.

    FourierField objects represent functions given by their Fourier coefficients
    on a given doubly-periodic square domain. They define arithmetic operations
    in real and spectral space, antialiasing when necessary, and can also be
    resampled to different grid resolutions.

    """

    # Prevent numpy from broadcasting instead of using arithmetic defined here.
    __array_ufunc__ = None

    def __init__(self, hats, p):
        """
        Initializes a FourierField given Fourier coefficients and a domain size.

        Note that in general FourierField objects should be created using one of
        the class constructors defined below, which in turn call this method.
        This method also checks that the grid has an odd number of points, to
        prevent differentiation errors from the unmatched mode.

        Parameters
        ----------
        hats : np.ndarray
            Two-dimensional array of Fourier coefficients, in the order used by
            numpy FFT routines.
        p : float
            The length of a side of the domain.

        Raises
        ------
        ValueError
            If the grid does not have an odd number of points.

        """

        self.hats = hats
        self.p = p

        if self.m % 2 == 0:
            raise ValueError('grid size must be odd')

    def __add__(self, other):
        """
        Adds one FourierField to another or to a real function.

        Parameters
        ----------
        other : int or float or complex or np.ndarray or FourierField
            Value to add to. If a numeric type or array, other represents a
            function by its values in real space, and addition is performed in
            real space after an IFFT. If a FourierField, other represents a
            function by its Fourier coefficients, and linearity of the FFT is
            leveraged to perform the addition in spectral space.

        Returns
        -------
        field : FourierField
            Spectral representation of the sum of self and other.

        """

        if isinstance(other, (int, float, complex, np.ndarray)):
            return type(self).from_data(self.real + other, self.p)

        return type(self)(self.hats + other.hats, self.p)

    def __sub__(self, other):
        """Subtraction is just negation followed by addition."""
        return self + -other

    def __mul__(self, other):
        """
        Multiplies one FourierField by another or by numerical values.

        Parameters
        ----------
        other : int or float or complex or np.ndarray or FourierField
            Value to multiply by. If a numeric type or array, other represents a
            scaling factor or array of factors for the Fourier coefficients of
            self, and multiplication is performed in spectral space. If other is
            a FourierField, multiplication is performed in real space after an
            IFFT, with both fields temporarily resampled at higher resolution to
            avoid aliasing effects.

        Returns
        -------
        field : FourierField
            Spectral representation of the product of self and other.

        """

        if isinstance(other, (int, float, complex, np.ndarray)):
            return type(self)(other * self.hats, self.p)

        m_aa = 3 * (self.m + 1) // 2
        if m_aa % 2 == 0:
            m_aa = m_aa + 1

        data = self.resample(m_aa).real * other.resample(m_aa).real

        return type(self).from_real(data, self.p).resample(self.m)

    def __rmul__(self, other):
        """Makes FourierField multiplication is always commutative."""
        return self * other

    def __neg__(self):
        """Allows FourierField objects to be negated directly."""
        return -1 * self

    @property
    def m(self):
        """The number of nodes in the spectral representation."""
        return self.hats.shape[0]

    @property
    def real(self):
        """The real-space representation of the function, computed via IFFT."""
        return utils.ifft(self.hats)

    def resample(self, m):
        """
        Resample a FourierField to a new spatial resolution.

        FourierField objects can be resampled to finer resolutions, in which
        case the array of coefficients is padded with zeros, or to coarser
        resolutions, in which case the coefficients are truncated.

        Parameters
        ----------
        m : int
            The new spatial resolution.

        Returns
        -------
        field : FourierField
            The spectral representation of the function resampled at the new
            spatial resolution.

        """

        if m == self.m:
            return self

        c = (min(m, self.m) + 1) // 2
        slices = [slice(c), slice(-(c - 1), None)]

        hats = np.zeros((m, m), dtype=self.hats.dtype)
        for idx, jdx in product(slices, slices):
            hats[idx, jdx] = self.hats[idx, jdx]

        hats = ((m / self.m) ** 2) * hats

        return type(self)(hats, self.p)

    @classmethod
    def from_func(cls, func, m, p=1):
        """
        Construct a FourierField from a real-valued function.

        Parameters
        ----------
        func : callable
            Real-valued function of two spatial coordinates. Should have the
            signature func(x, y), where x and y are numpy arrays.
        m : int
            The desired grid resolution.
        p : float, optional
            The size of the square domain. Default is 1, in which case the
            domain is the doubly-periodic unit square.

        Returns
        -------
        field : FourierField
            Spectral representation of the function applied to [0, p]^2.

        """

        x, y = utils.make_grid(m, p)
        hats = utils.fft(func(x, y))

        return cls(hats, p)

    @classmethod
    def from_real(cls, real, p=1):
        """
        Construct a FourierField from gridded real data on a square domain.

        Parameters
        ----------
        real : np.ndarray
            Real-space function values on a square grid.
        p : float, optional
            The size of the square domain. Default is 1, in which case the
            domain is the doubly-periodic unit square.

        Returns
        -------
        field : FourierField
            Spectral representation of the function.

        """

        return cls(utils.fft(real), p)



import numpy as np

from whirly.integrators import IFRK4Integrator

def test_fourth_order():
    nonlinear = lambda y: y ** 2
    y_func = lambda t: np.exp(t + 1) / (1 - np.exp(t + 1))

    taus, errors = [0.01, 0.001], []
    for tau in taus:
        integrator = IFRK4Integrator(tau, 1, nonlinear)
        y = y_func(0)

        n_steps = int(1 / tau)
        for _ in range(n_steps):
            y = integrator.step(y)

        errors.append(abs(y - y_func(1)))

    p = np.log(errors[0] / errors[1]) / np.log(taus[0] / taus[1])

    assert round(p) == 4
